/***********************************************************************
 *
 * Copyright (C) 2008-2009 Graeme Gott <graeme@gottcode.org>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 ***********************************************************************/

#include "data_model.h"

#include <QApplication>
#include <QMessageBox>
#include <QPalette>

#include <cmath>

/*****************************************************************************/

namespace {
	float hours(int seconds) {
		return std::floor(seconds / 360.f) * 0.1f;
	}

	QString hoursToolTip(int seconds) {
		int hours = std::floor(seconds / 3600.f);
		seconds -= (hours * 3600);
		int minutes = std::floor(seconds / 60.f);
		seconds -= (minutes * 60);
		QString hours_str = (hours == 1) ? DataModel::tr("1 hour") : DataModel::tr("%1 hours").arg(hours);
		QString minutes_str = (minutes == 1) ? DataModel::tr("1 minute") : DataModel::tr("%1 minutes").arg(minutes);
		QString seconds_str = (seconds == 1) ? DataModel::tr("1 second") : DataModel::tr("%1 seconds").arg(seconds);
		if (hours > 0) {
			return QString("%1, %2, %3").arg(hours_str).arg(minutes_str).arg(seconds_str);
		} else if (minutes > 0) {
			return QString("%1, %2").arg(minutes_str).arg(seconds_str);
		} else if (seconds) {
			return seconds_str;
		} else {
			return QString();
		}
	}
}

/*****************************************************************************/

DataModel::DataModel(QObject* parent)
: QAbstractTableModel(parent) {
}

/*****************************************************************************/

bool DataModel::add(const QDateTime& start, const QDateTime& stop) {
	if (stop < start) {
		return false;
	}

	if (start.date() == stop.date()) {
		return add(start.date(), start.time(), stop.time(), false, QString());
	} else {
		return add(start.date(), start.time(), QTime(23, 59, 59), false, QString()) && add(stop.date(), QTime(0, 0, 0), stop.time(), false, QString());
	}
}

/*****************************************************************************/

bool DataModel::add(const QDate& date, const QTime& start, const QTime& stop, bool billed, const QString& note) {
	if (stop < start) {
		return false;
	}

	// Find session position
	int pos = 0;
	foreach (const Session& session, m_data) {
		if (date < session.date || (date == session.date && start < session.start)) {
			break;
		}
		pos++;
	}

	// Prevent intersecting sessions
	if (pos > 0) {
		const Session& session = m_data.at(pos - 1);
		if (QDateTime(date, start) < QDateTime(session.date, session.stop)) {
			return false;
		}
	}
	if (pos < m_data.count()) {
		const Session& session = m_data.at(pos);
		if (QDateTime(date, stop) > QDateTime(session.date, session.start)) {
			return false;
		}
	}

	// Insert session
	beginInsertRows(QModelIndex(), pos, pos);
	int duration = start.secsTo(stop);
	int daily = duration;
	int weekly = duration;
	int monthly = duration;
	int previous = pos - 1;
	if (pos > 0 && previous < m_data.count() && !m_data.at(previous).billed) {
		const Session& session = m_data.at(previous);
		if (date == session.date) {
			daily += session.daily;
		}
		if (date.weekNumber() == session.date.weekNumber()) {
			weekly += session.weekly;
		}
		if (date.month() == session.date.month()) {
			monthly += session.monthly;
		}
	}
	Session session = { date, start, stop, daily, weekly, monthly, billed, note };
	m_data.insert(pos, session);
	endInsertRows();

	// Update successive billed sessions
	for (int i = 0; i < m_billed.count(); ++i) {
		int& session = m_billed[i];
		if (session >= pos) {
			session++;
		}
	}

	// Set billed status
	if (billed) {
		setBilled(pos, true);
	}

	// Increase totals for successive sessions
	updateTotals();

	return true;
}

/*****************************************************************************/

bool DataModel::edit(int pos, const QDate& date, const QTime& start, const QTime& stop, const QString& note) {
	if (stop < start) {
		return false;
	}

	Q_ASSERT(pos < m_data.count());
	const Session& session = m_data.at(pos);
	QDate old_date = session.date;
	QTime old_start = session.start;
	QTime old_stop = session.stop;
	bool billed = session.billed;

	// Replace session
	remove(pos);
	if (add(date, start, stop, billed, note) == false) {
		add(old_date, old_start, old_stop, billed, note);
		return false;
	} else {
		return true;
	}
}

/*****************************************************************************/

bool DataModel::remove(int pos) {
	if (pos >= m_data.count()) {
		return false;
	}

	// Remove session
	beginRemoveRows(QModelIndex(), pos, pos);
	m_data.removeAt(pos);
	endRemoveRows();

	// Update successive billed sessions
	m_billed.removeAll(pos);
	for (int i = 0; i < m_billed.count(); ++i) {
		int& session = m_billed[i];
		if (session > pos) {
			session--;
		}
	}

	return true;
}

/*****************************************************************************/

void DataModel::setBilled(int pos, bool billed) {
	Q_ASSERT(pos < m_data.count());

	if (!billed) {
		m_billed.removeAll(pos);
	} else {
		Q_ASSERT(!m_billed.contains(pos));
		m_billed.append(pos);
		qSort(m_billed);
	}
	m_data[pos].billed = billed;

	updateTotals();
}

/*****************************************************************************/

void DataModel::toXml(QXmlStreamWriter& xml) const {
	foreach (const Session& session, m_data) {
		xml.writeEmptyElement("session");
		xml.writeAttribute("date", session.date.toString(Qt::ISODate));
		xml.writeAttribute("start", session.start.toString(Qt::ISODate));
		xml.writeAttribute("stop", session.stop.toString(Qt::ISODate));
		xml.writeAttribute("billed", QString::number(session.billed));
		xml.writeAttribute("note", session.note);
	}
}

/*****************************************************************************/

int DataModel::rowCount(const QModelIndex& parent) const {
	if (!parent.isValid()) {
		return m_data.count();
	} else {
		return 0;
	}
}

/*****************************************************************************/

int DataModel::columnCount(const QModelIndex& parent) const {
	if (!parent.isValid()) {
		return 8;
	} else {
		return 0;
	}
}

/*****************************************************************************/

QVariant DataModel::data(const QModelIndex& index, int role) const {
	Q_ASSERT(index.isValid());
	QVariant result;

	const Session& session = m_data.at(index.row());
	switch (role) {
	case Qt::DisplayRole:
		switch (index.column()) {
		case 0:
			result = session.date.toString("M/d/yy");
			break;
		case 1:
			result = session.start.toString("h:mm A");
			break;
		case 2:
			result = session.stop.toString("h:mm A");
			break;
		case 3:
			result = hours(session.daily);
			break;
		case 4:
			result = hours(session.weekly);
			break;
		case 5:
			result = hours(session.monthly);
			break;
		case 7:
			result = session.note;
			break;
		default:
			break;
		}
		break;

	case Qt::ToolTipRole:
		switch (index.column()) {
		case 3:
			result = hoursToolTip(session.daily);
			break;
		case 4:
			result = hoursToolTip(session.weekly);
			break;
		case 5:
			result = hoursToolTip(session.monthly);
			break;
		default:
			break;
		}
		break;

	case Qt::TextAlignmentRole:
		if (index.column() != 7) {
			result = static_cast<int>(Qt::AlignRight | Qt::AlignVCenter);
		}
		break;

	case Qt::ForegroundRole:
		if (isBilled(index.row())) {
			result = QApplication::palette().color(QPalette::Disabled, QPalette::Text);
		}
		break;

	case Qt::CheckStateRole:
		if (index.column() == 6) {
			result = m_data.at(index.row()).billed ? Qt::Checked : Qt::Unchecked;
		}
		break;

	case Qt::EditRole:
		Q_ASSERT(!isBilled(index.row()));
		switch (index.column()) {
		case 0:
			result = session.date;
			break;
		case 1:
			result = session.start;
			break;
		case 2:
			result = session.stop;
			break;
		case 7:
			result = session.note;
			break;
		default:
			break;
		}
		break;

	default:
		break;
	}
	return result;
}

/*****************************************************************************/

Qt::ItemFlags DataModel::flags(const QModelIndex& index) const {
	if (index.column() == 6) {
		return QAbstractTableModel::flags(index) |  Qt::ItemIsUserCheckable;
	} else if (isBilled(index.row()) || (index.column() > 2 && index.column() < 6)) {
		return QAbstractTableModel::flags(index);
	} else {
		return QAbstractTableModel::flags(index) | Qt::ItemIsEditable;
	}
}

/*****************************************************************************/

QVariant DataModel::headerData(int section, Qt::Orientation orientation, int role) const {
	if (orientation != Qt::Horizontal) {
		return QVariant();
	}

	if (role == Qt::DisplayRole) {
		switch (section) {
		case 0:
			return tr("Date");
		case 1:
			return tr("Start");
		case 2:
			return tr("Stop");
		case 3:
			return tr("Daily");
		case 4:
			return tr("Weekly");
		case 5:
			return tr("Monthly");
		case 6:
			return tr("Billed");
		case 7:
			return tr("Note");
		default:
			return QVariant();
		}
	} else {
		return QVariant();
	}
}

/*****************************************************************************/

bool DataModel::setData(const QModelIndex& index, const QVariant& value, int role) {
	if (!isBilled(index.row()) && role == Qt::EditRole) {
		const Session& session = m_data.at(index.row());
		QDate date = session.date;
		QTime start = session.start;
		QTime stop = session.stop;
		QString note = session.note;
		switch (index.column()) {
		case 0:
			date = value.toDate();
			break;
		case 1:
			start = value.toTime();
			break;
		case 2:
			stop = value.toTime();
			break;
		case 7:
			note = value.toString();
			break;
		default:
			break;
		}

		if (edit(index.row(), date, start, stop, note)) {
			return true;
		} else {
			QMessageBox::warning(0, tr("Error"), tr("Session conflicts with other sessions."));
			return false;
		}
	} else if (index.column() == 6 && role == Qt::CheckStateRole) {
		setBilled(index.row(), value.toInt() == Qt::Checked);
		emit dataChanged(index, index);
		return true;
	} else {
		return QAbstractTableModel::setData(index, value, role);
	}
}

/*****************************************************************************/

void DataModel::updateTotals() {
	const Session& first = m_data.at(0);
	bool billed = first.billed;
	int daily = first.daily;
	int weekly = first.weekly;
	int monthly = first.monthly;

	for (int i = 1; i < m_data.count(); ++i) {
		const QDate& date = m_data.at(i - 1).date;
		Session& session = m_data[i];
		int duration = QDateTime(session.date, session.stop).toTime_t() - QDateTime(session.date, session.start).toTime_t();
		if (billed) {
			daily = weekly = monthly = duration;
		} else {
			if (session.date == date) {
				daily += duration;
			} else {
				daily = duration;
			}
			if (session.date.weekNumber() == date.weekNumber()) {
				weekly += duration;
			} else {
				weekly = duration;
			}
			if (session.date.month() == date.month()) {
				monthly += duration;
			} else {
				monthly = duration;
			}
		}
		session.daily = daily;
		session.weekly = weekly;
		session.monthly = monthly;
		billed = session.billed;
	}

	emit dataChanged(index(0, 0), index(m_data.count() - 1, 6));
}

/*****************************************************************************/
