/***********************************************************************
 *
 * Copyright (C) 2008-2009 Graeme Gott <graeme@gottcode.org>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 ***********************************************************************/

#include "window.h"

#include "data_model.h"
#include "date_editor.h"
#include "project.h"
#include "session_dialog.h"
#include "time_editor.h"

#include <QApplication>
#include <QDateTime>
#include <QFile>
#include <QHBoxLayout>
#include <QHeaderView>
#include <QInputDialog>
#include <QItemDelegate>
#include <QItemEditorFactory>
#include <QLabel>
#include <QMessageBox>
#include <QMetaProperty>
#include <QPushButton>
#include <QSettings>
#include <QSplitter>
#include <QTimer>
#include <QTreeView>
#include <QTreeWidget>
#include <QTreeWidgetItem>
#include <QVBoxLayout>
#include <QXmlStreamReader>

/*****************************************************************************/

namespace {
	class Delegate : public QItemDelegate {
	public:
		Delegate(QObject* parent = 0)
		: QItemDelegate(parent) {
		}

		QWidget* createEditor(QWidget* parent, const QStyleOptionViewItem& option, const QModelIndex& index) const {
			if (index.column() == 0) {
				return QItemDelegate::createEditor(parent, option, index);
			} else {
				return 0;
			}
		}
	};
}

/*****************************************************************************/

Window::Window(QWidget* parent)
: QWidget(parent), m_active_project(0), m_active_model(0), m_active_timers(0) {
	setWindowTitle(tr("Kapow Punch Clock"));
	setWindowIcon(QIcon(":/kapow.png"));

	m_display = new QLabel(tr("00:00:00"), this);
	QFont font = m_display->font();
	font.setPointSize(32);
	m_display->setFont(font);

	m_current_time = QDateTime::currentDateTime();
	m_timer = new QTimer(this);
	m_timer->setInterval(1000);
	connect(m_timer, SIGNAL(timeout()), this, SLOT(updateTime()));
	m_timer->start();

	QTimer* save_timer = new QTimer(this);
	save_timer->setInterval(30000);
	connect(save_timer, SIGNAL(timeout()), this, SLOT(save()));
	save_timer->start();

	m_start = new QPushButton(tr("Start"), this);
	connect(m_start, SIGNAL(clicked()), this, SLOT(start()));

	m_stop = new QPushButton(tr("Stop"), this);
	m_stop->hide();
	connect(m_stop, SIGNAL(clicked()), this, SLOT(stop()));

	m_cancel = new QPushButton(tr("Cancel"), this);
	m_cancel->setEnabled(false);
	connect(m_cancel, SIGNAL(clicked()), this, SLOT(cancel()));

	// Create projects
	m_projects = new QTreeWidget(this);
	m_projects->setAlternatingRowColors(true);
	m_projects->setAutoExpandDelay(500);
	m_projects->setDragEnabled(true);
	m_projects->setDragDropMode(QAbstractItemView::InternalMove);
	m_projects->setHeaderLabels(QStringList() << tr("Project") << tr("Timer"));
	m_projects->setItemDelegate(new Delegate(m_projects));
	m_projects->setSelectionMode(QAbstractItemView::SingleSelection);
	m_projects->setSelectionBehavior(QAbstractItemView::SelectRows);
	m_projects->setSortingEnabled(true);
	m_projects->setVerticalScrollMode(QAbstractItemView::ScrollPerPixel);
	m_projects->sortByColumn(0, Qt::AscendingOrder);
	m_projects->header()->setClickable(false);
	m_projects->header()->setMovable(false);
	m_projects->header()->setResizeMode(0, QHeaderView::Stretch);
	m_projects->header()->setResizeMode(1, QHeaderView::Fixed);
	m_projects->header()->setSortIndicatorShown(false);
	m_projects->header()->setStretchLastSection(false);
	m_projects->setColumnWidth(1, fontMetrics().averageCharWidth() * 10);
	connect(m_projects, SIGNAL(currentItemChanged(QTreeWidgetItem*, QTreeWidgetItem*)), this, SLOT(projectActivated(QTreeWidgetItem*)));

	m_add_project = new QPushButton(tr("Add Project"), this);
	m_add_project->setIconSize(QSize(16,16));
	connect(m_add_project, SIGNAL(clicked()), this, SLOT(addProject()));

	m_remove_project = new QPushButton(tr("Remove Project"), this);
	m_remove_project->setIconSize(QSize(16,16));
	m_remove_project->setEnabled(false);
	connect(m_remove_project, SIGNAL(clicked()), this, SLOT(removeProject()));

	QVBoxLayout* project_layout = new QVBoxLayout;
	project_layout->setMargin(0);
	project_layout->setSpacing(0);
	project_layout->addWidget(m_add_project);
	project_layout->addWidget(m_remove_project);

	// Create details
	QItemEditorFactory* factory = new QItemEditorFactory;
	factory->registerEditor(QVariant::Date, new QStandardItemEditorCreator<DateEditor>());
	factory->registerEditor(QVariant::Time, new QStandardItemEditorCreator<TimeEditor>());
	factory->registerEditor(QVariant::String, new QStandardItemEditorCreator<QLineEdit>());
	QItemEditorFactory::setDefaultFactory(factory);

	m_details = new QTreeView(this);
	m_details->setRootIsDecorated(false);
	m_details->setSelectionMode(QAbstractItemView::SingleSelection);
	m_details->setSelectionBehavior(QAbstractItemView::SelectRows);
	m_details->setVerticalScrollMode(QAbstractItemView::ScrollPerPixel);
	m_details->header()->setClickable(false);
	m_details->header()->setMovable(false);
	m_details->header()->setResizeMode(QHeaderView::ResizeToContents);
	connect(m_details, SIGNAL(pressed(const QModelIndex&)), this, SLOT(sessionPressed(const QModelIndex&)));

	m_add_session = new QPushButton(tr("Add Session"), this);
	m_add_session->setIconSize(QSize(16,16));
	connect(m_add_session, SIGNAL(clicked()), this, SLOT(addSession()));

	m_remove_session = new QPushButton(tr("Remove Session"), this);
	m_remove_session->setIconSize(QSize(16,16));
	m_remove_session->setEnabled(false);
	connect(m_remove_session, SIGNAL(clicked()), this, SLOT(removeSession()));

	QVBoxLayout* session_layout = new QVBoxLayout;
	session_layout->setMargin(0);
	session_layout->setSpacing(0);
	session_layout->addWidget(m_add_session);
	session_layout->addWidget(m_remove_session);

	// Create contents splitter
	m_contents = new QSplitter(this);
	m_contents->addWidget(m_projects);
	m_contents->addWidget(m_details);
	m_contents->setStretchFactor(0, 0);
	m_contents->setStretchFactor(1, 1);

	// Load details of all projects
	DataModel* model = 0;
	QTreeWidgetItem* item = 0;
	QTreeWidgetItem* current = 0;
	QFile file("data.xml");
	if (file.open(QIODevice::ReadOnly | QIODevice::Text)) {
		QXmlStreamReader xml(&file);
		while (!xml.atEnd()) {
			xml.readNext();
			if (xml.isStartElement()) {
				QXmlStreamAttributes attributes = xml.attributes();
				if (xml.name() == QLatin1String("session")) {
					QDate date = QDate::fromString(attributes.value("date").toString(), Qt::ISODate);
					QTime start = QTime::fromString(attributes.value("start").toString(), Qt::ISODate);
					QTime stop = QTime::fromString(attributes.value("stop").toString(), Qt::ISODate);
					bool billed = attributes.value("billed").toString().toInt();
					QString note = attributes.value("note").toString();
					model->add(date, start, stop, billed, note);
				} else if (xml.name() == QLatin1String("project")) {
					m_projects->blockSignals(true);
					QString name = attributes.value("name").toString();
					Project* project = 0;
					if (item == 0) {
						project = new Project(m_projects, name);
					} else {
						project = new Project(item, name);
					}
					project->setExpanded(attributes.value("expanded") == "1");
					current = (attributes.value("current") == "1") ? project : current;
					item = project;
					model = project->model();
					m_projects->blockSignals(false);
				} else if (xml.name() == QLatin1String("autosave")) {
					QDateTime start = QDateTime::fromString(attributes.value("start").toString(), Qt::ISODate);
					QDateTime stop = QDateTime::fromString(attributes.value("stop").toString(), Qt::ISODate);
					model->add(start, stop);
				}
			} else if (xml.isEndElement() && xml.name() == QLatin1String("project")) {
				item = item->parent();
			}
		}
	}
	m_start->setEnabled(m_projects->topLevelItemCount() > 0);

	// Load last used project
	if (current == 0) {
		current = m_projects->topLevelItem(0);
	}
	m_projects->setCurrentItem(current);

	// Lay out window
	QVBoxLayout* session_buttons = new QVBoxLayout;
	session_buttons->setMargin(0);
	session_buttons->setSpacing(0);
	session_buttons->addWidget(m_start);
	session_buttons->addWidget(m_stop);
	session_buttons->addWidget(m_cancel);

	QHBoxLayout* details_layout = new QHBoxLayout;
	details_layout->setMargin(0);
	details_layout->addLayout(project_layout);
	details_layout->addStretch();
	details_layout->addWidget(m_display);
	details_layout->addLayout(session_buttons);
	details_layout->addStretch();
	details_layout->addLayout(session_layout);

	QVBoxLayout* layout = new QVBoxLayout(this);
	layout->addLayout(details_layout);
	layout->addWidget(m_contents);

	// Restore window geometry
	resize(900, 600);
	QSettings settings;
	restoreGeometry(settings.value("WindowGeometry").toByteArray());
	m_contents->restoreState(settings.value("SplitterSizes").toByteArray());

	// Add tray icon
	m_tray_icon = new QSystemTrayIcon(QIcon(":/kapow.png"), this);
	updateTrayIcon();
	connect(m_tray_icon, SIGNAL(activated(QSystemTrayIcon::ActivationReason)), this, SLOT(trayIconActivated(QSystemTrayIcon::ActivationReason)));
	m_tray_icon->show();
}

/*****************************************************************************/

void Window::closeEvent(QCloseEvent* event) {
	QSettings settings;
	settings.setValue("WindowGeometry", saveGeometry());
	settings.setValue("SplitterSizes", m_contents->saveState());
	save();
	QWidget::closeEvent(event);
}

/*****************************************************************************/

void Window::start() {
	m_active_project->start(m_current_time);
	m_active_timers++;
	updateDetails();
}

/*****************************************************************************/

void Window::stop() {
	m_active_project->stop(m_current_time);
	m_active_timers--;
	updateDetails();
}

/*****************************************************************************/

void Window::cancel() {
	if (QMessageBox::question(this, tr("Question"), tr("Cancel this session?"), QMessageBox::Yes | QMessageBox::No, QMessageBox::No) == QMessageBox::Yes) {
		m_active_project->stop();
		m_active_timers--;
		updateDetails();
	}
}

/*****************************************************************************/

void Window::updateTime() {
	m_current_time = QDateTime::currentDateTime();
	int count = m_projects->topLevelItemCount();
	for (int i = 0; i < count; ++i) {
		Project* project = dynamic_cast<Project*>(m_projects->topLevelItem(i));
		if (project) {
			project->updateTime(m_current_time);
		}
	}
	if (m_active_project) {
		updateDisplay();
	}
}

/*****************************************************************************/

void Window::addProject() {
	bool ok;
	QString name;
	forever {
		name = QInputDialog::getText(this, tr("Add Project"), tr("Enter project name:"), QLineEdit::Normal, name, &ok);
		if (ok && !name.isEmpty()) {
			if (m_projects->findItems(name, Qt::MatchExactly).isEmpty()) {
				m_start->setEnabled(true);
				Project* project = new Project(m_projects, name);
				m_projects->setCurrentItem(project);
				return;
			} else {
				QMessageBox::warning(this, tr("Error"), tr("A project with that name already exists."));
			}
		} else {
			break;
		}
	}
}

/*****************************************************************************/

void Window::editProject() {
	m_projects->edit(m_projects->currentIndex());
}

/*****************************************************************************/

void Window::removeProject() {
	if (QMessageBox::question(this, tr("Question"), tr("Remove selected project?"), QMessageBox::Yes | QMessageBox::No, QMessageBox::No) == QMessageBox::Yes) {
		m_remove_project->setEnabled(false);
		m_add_session->setEnabled(false);
		m_remove_session->setEnabled(false);
		removeProject(m_projects->currentItem());
	}
}

/*****************************************************************************/

void Window::projectActivated(QTreeWidgetItem* item) {
	Project* project = dynamic_cast<Project*>(item);
	if (!project) {
		return;
	}

	if (m_active_model) {
		disconnect(m_active_model, SIGNAL(rowsInserted(const QModelIndex&, int, int)), this, SLOT(sessionsInserted(const QModelIndex&, int, int)));
	}
	m_active_project = project;
	m_active_model = m_active_project->model();
	connect(m_active_model, SIGNAL(rowsInserted(const QModelIndex&, int, int)), this, SLOT(sessionsInserted(const QModelIndex&, int, int)));

	updateDisplay();
	updateSessionButtons();
	m_details->setModel(m_active_model);
	m_remove_project->setEnabled(true);
	m_add_session->setEnabled(true);
	m_remove_session->setEnabled(false);
}

/*****************************************************************************/

void Window::sessionPressed(const QModelIndex& index) {
	bool enabled = !m_active_model->isBilled(index.row());
	m_remove_session->setEnabled(enabled);
}

/*****************************************************************************/

void Window::sessionsInserted(const QModelIndex&, int, int end) {
	m_details->setCurrentIndex(m_active_model->index(end, 0));
}

/*****************************************************************************/

void Window::addSession() {
	Q_ASSERT(m_active_model != 0);
	SessionDialog dialog(this);
	forever {
		if (dialog.exec() == QDialog::Accepted) {
			if (m_active_model->add(dialog.date(), dialog.start(), dialog.stop(), false, dialog.note())) {
				return;
			}
			QMessageBox::warning(this, tr("Error"), tr("Session conflicts with other sessions."));
		} else {
			break;
		}
	}
}

/*****************************************************************************/

void Window::editSession() {
	m_details->edit(m_details->currentIndex());
}

/*****************************************************************************/

void Window::removeSession() {
	Q_ASSERT(m_active_model != 0);
	if (QMessageBox::question(this, tr("Question"), tr("Remove selected session?"), QMessageBox::Yes | QMessageBox::No, QMessageBox::No) == QMessageBox::Yes) {
		m_active_model->remove(m_details->currentIndex().row());
	}
}

/*****************************************************************************/

void Window::save() {
	QFile file("data.xml");
	file.open(QIODevice::WriteOnly | QIODevice::Text);
	QXmlStreamWriter xml(&file);
	xml.setAutoFormatting(true);
	xml.writeStartDocument();
	xml.writeStartElement("kapow");
	int count = m_projects->topLevelItemCount();
	for (int i = 0; i < count; ++i) {
		Project* project = dynamic_cast<Project*>(m_projects->topLevelItem(i));
		if (project) {
			project->toXml(xml);
		}
	}
	xml.writeEndDocument();
}

/*****************************************************************************/

void Window::trayIconActivated(QSystemTrayIcon::ActivationReason reason) {
	if (reason == QSystemTrayIcon::Trigger) {
		setVisible(!isVisible());
	}
}

/*****************************************************************************/

void Window::removeProject(QTreeWidgetItem* item) {
	int count = item->childCount();
	for (int i = 0; i < count; ++i) {
		removeProject(item->child(i));
	}
	Project* project = dynamic_cast<Project*>(item);
	if (project && project->model() == m_active_model)  {
		m_active_model = 0;
	}
	delete item;
	item = 0;
	m_start->setDisabled(m_projects->topLevelItemCount() == 0);
}

/*****************************************************************************/

void Window::updateDetails() {
	Q_ASSERT(m_active_project != 0);
	updateDisplay();
	updateSessionButtons();
	updateTrayIcon();
}

/*****************************************************************************/

void Window::updateDisplay() {
	QString time = m_active_project->time();
	m_display->setText(!time.isEmpty() ? time : "00:00:00");
}

/*****************************************************************************/

void Window::updateSessionButtons() {
	if (!m_active_project->time().isEmpty()) {
		m_start->hide();
		m_stop->show();
		m_cancel->setEnabled(true);
	} else {
		m_stop->hide();
		m_start->show();
		m_cancel->setEnabled(false);
	}
}

/*****************************************************************************/

void Window::updateTrayIcon() {
	QString tip;
	switch (m_active_timers) {
	case 0:
		tip = tr("Kapow - No timers running");
		break;
	case 1:
		tip = tr("Kapow - 1 timer running");
		break;
	default:
		tip = tr("Kapow - %1 timers running").arg(m_active_timers);
		break;
	}
	m_tray_icon->setToolTip(tip);
}

/*****************************************************************************/
